@@ -1,5 +1,4 @@
-import copy
-from typing import Union
+from typing import Optional, Union
 
 TABLE_PROPERTIES = ('widths', 'borders', 'fills')
 PARAGRAPH_PROPERTIES = (
@@ -118,16 +117,43 @@ class PDFContent:
             'borders': [{'pos': 'h0,1,3;:', 'width': 2, 'color': 0}]
         }
 
-    * A content box that can be a dict like the one being explained here, and can
+    * A content box that can be a dict like the one being explained here, and
       can contain other elements inside it recursively. This can be used to
       insert a new section with more columns (for example a 2 columns content
       box, inside another 2 columns content box).
 
+    * A group element that is a list of paragraphs, images or tables that should
+      be placed all in the same page. This can be used for example to place an
+      image with a description, with the guarantee that both will be in the same
+      page. Be careful though, because the group element should fit the
+      width and max height of the containing box, or else an error will be
+      raised. This can be "relaxed" by setting ``min_height`` property style in
+      the images inside the group. If an image does not have this property it
+      will take as much space as possible, and if it does it will be shrinked as
+      much as possible (without shrinking it beyond ``min_height``) to make the
+      other elements in the group fit in the available height. If there are more
+      than one images in the group with ``min_height`` style property they will
+      be shrinked together proportionally. If you want to ensure that some image
+      will be shrinked until its ``min_height``, use the ``shrink`` style
+      property.
+      Here is an example of a group element:
+
+      .. code-block:: python
+
+        {
+            "style": {"margin_left": 80, "margin_right": 80},
+            "group": [
+                {"image": "tests/image_test.jpg", "min_height": 200},
+                {".": "Figure 1: Description of figure 1"}
+            ]
+        }
+
     Each element in the content box can have margins to keep it separated from
     the other elements, and these margins can be set inside the ``style`` dict
-    of the content box dict with the following keys: ``margin_top``, ``margin_left``,
-    ``margin_bottom`` and ``margin_right``. Default value for all of them is 0,
-    except for ``margin_bottom`` that have a default value of 5.
+    of the content box dict with the following keys: ``margin_top``,
+    ``margin_left``, ``margin_bottom`` and ``margin_right``. Default value for
+    all of them is 0, except for ``margin_bottom`` that have a default value of
+    5.
 
     All of the children elements in the content box will inherit the
     the content box style.
@@ -232,6 +258,37 @@ class PDFContent:
         if ret == 'continue':
             self.finished = True
 
+    def get_state(self) -> dict:
+        """Method to get the current state of this content box. This can be used
+        later in method :meth:`pdfme.content.PDFContent.set_state` to restore
+        this state in this content box (like a checkpoint in a videogame).
+
+        Returns:
+            dict: a dict with the state of this content box.
+        """
+        return self.pdf_content_part.get_state()
+
+    def set_state(
+        self, section_element_index: int=None, section_delayed: list=None,
+        children_memory: list=None
+    ) -> None:
+        """Method to set the state of this content box.
+
+        The 3 arguments of this method define the current state of this content
+        box, and with this method you can change that state.
+
+        Args:
+            section_element_index (int, optional): the index of the current
+                element being added.
+            section_delayed (list, optional): a list of delayed elements, that
+                should be added before continuing with the rest of elements.
+            children_memory (list, optional): if the current element is in turn
+                a content box, this list says what the indexes of the nested
+                content boxes inside this content box are.
+        """
+        self.pdf_content_part.set_state(
+            section_element_index, section_delayed, children_memory
+        )
 class PDFContentPart:
     """Class that represent a content element.
 
@@ -328,7 +385,7 @@ class PDFContentPart:
         self.col_width = (width - cols_spaces) / self.cols_n
 
         self.element_index = self.section_element_index # current index
-        self.delayed = copy.deepcopy(self.section_delayed) # current delayed elements
+        self.delayed = copy(self.section_delayed) # current delayed elements
         self.will_reset = False
         self.resetting = False
         self.parts_index = len(self.p.parts)
@@ -337,6 +394,43 @@ class PDFContentPart:
         self.minim_diff = None
         self.minim_forward = None
 
+    def get_state(self) -> dict:
+        """Method to get the current state of this content box. This can be used
+        later in method :meth:`pdfme.content.PDFContentPart.set_state` to
+        restore this state in this content box (like a checkpoint in a
+        videogame).
+
+        Returns:
+            dict: a dict with the state of this content box.
+        """
+        return {
+            'section_element_index': self.section_element_index,
+            'section_delayed': copy(self.section_delayed),
+            'children_memory': copy(self.children_memory)
+        }
+
+    def set_state(
+        self, section_element_index: int=None, section_delayed: list=None,
+        children_memory: list=None
+    ) -> None:
+        """Method to set the state of this content box part.
+
+        The arguments of this method define the current state of this content
+        box part, and with this method you can change that state.
+
+        Args:
+            section_element_index (int, optional): the index of the current
+                element being added.
+            section_delayed (list, optional): a list of delayed elements, that
+                should be added before continuing with the rest of elements.
+            children_memory (list, optional): if the current element is in turn
+                a content box, this list says what the indexes of the nested
+                content boxes inside this content box are.
+        """
+        self.section_element_index = section_element_index
+        self.section_delayed = copy(section_delayed)
+        self.children_memory = copy(children_memory)
+
     def add_delayed(self) -> str:
         '''Function to add the delayed elements to the rectangle.
 
@@ -351,19 +445,19 @@ class PDFContentPart:
         '''
         n = 0
         while n < len(self.delayed):
-            ret = self.process(copy.deepcopy(self.delayed[n]), False)
+            ret = self.process(copy(self.delayed[n]), False)
             if ret in ['interrupt', 'break', 'partial_next']:
                 return ret
 
             if ret.get('delayed'):
-                self.delayed[n] = copy.deepcopy(ret['delayed'])
+                self.delayed[n] = copy(ret['delayed'])
             else:
                 self.delayed.pop(n)
 
             if ret.get('next', False):
                 return 'next'
 
-            if ret.get('image_flow', False):
+            if ret.get('flow', False):
                 n += 1
 
         if (
@@ -513,7 +607,7 @@ class PDFContentPart:
         is resetting.
 
         Returns:
-            True if this element is the las element of an ancestor that is
+            True if this element is the last element of an ancestor that is
             resetting.
         """
         parent = self.parent
@@ -547,7 +641,7 @@ class PDFContentPart:
             True if resetting process should continue or False if this process
             is done.
         """
-        if self.minim_diff_last and self.minim_diff_last - self.minim_diff < 10:
+        if self.minim_diff_last and self.minim_diff_last - self.minim_diff < 1:
             if self.minim_forward:
                 self.minim_diff *= 2
             else:
@@ -571,7 +665,7 @@ class PDFContentPart:
         self.resetting = True
 
         self.element_index = self.section_element_index
-        self.delayed = copy.deepcopy(self.section_delayed)
+        self.delayed = copy(self.section_delayed)
 
         return True
 
@@ -611,17 +705,17 @@ class PDFContentPart:
                 return 'break'
             else:
                 self.section_element_index = self.element_index
-                self.section_delayed = copy.deepcopy(self.delayed)
+                self.section_delayed = copy(self.delayed)
                 new_index = {
                     'index': self.element_index,
-                    'delayed': copy.deepcopy(self.delayed)
+                    'delayed': copy(self.delayed)
                 }
                 if children_memory is None:
                     self.children_memory = []
                     new_children_memory = [new_index]
                 else:
-                    self.children_memory = copy.deepcopy(children_memory)
-                    new_children_memory = copy.deepcopy(children_memory)
+                    self.children_memory = copy(children_memory)
+                    new_children_memory = copy(children_memory)
                     new_children_memory.append(new_index)
 
                 if self.is_root:
@@ -641,7 +735,7 @@ class PDFContentPart:
             self.y = self.min_y
 
             if len(self.delayed) > 0 and children_memory is not None:
-                self.partial_children_memory = copy.deepcopy(children_memory)
+                self.partial_children_memory = copy(children_memory)
                 return 'partial_next'
 
             return 'retry' if children_memory is None else \
@@ -673,7 +767,6 @@ class PDFContentPart:
             self.y -= self.last_bottom
 
         self.max_height = max(0, self.y - self.max_y)
-        self.last_bottom = s.get('margin_bottom', 0)
 
     def add_top_margin(self, style: dict) -> None:
         """Function that adds the top margin of the current child element.
@@ -687,6 +780,19 @@ class PDFContentPart:
         else:
             self.y -=  style.get('margin_top', 0)
 
+    def parse_element(self, element: ProcessElement) -> dict:
+        if not isinstance(element, (dict, str, list, tuple)):
+            return str(element)
+        if isinstance(element, (str, list, tuple)):
+            return {'.': element}
+
+        if not isinstance(element, dict):
+            raise TypeError(
+                'Elements must be of type dict, str, list or tuple:{}'
+                .format(element)
+            )
+        return element
+
     def process(self, element: ProcessElement, last: bool=False) -> StrOrDict:
         '''Function to add a single child element to the rectangle.
 
@@ -703,28 +809,10 @@ class PDFContentPart:
             last (bool): Wheter or not this is the last child element of the
                 list of child elements of this element.
         '''
-        if not isinstance(element, (dict, str, list, tuple)):
-            element = str(element)
-        if isinstance(element, (str, list, tuple)):
-            element = {'.': element}
-
-        if not isinstance(element, dict):
-            raise TypeError(
-                'Elements must be of type dict, str, list or tuple:{}'
-                .format(element)
-            )
+        element = self.parse_element(element)
+        style, element_style = self.get_element_styles(element, self.style)
 
         keys = [key for key in element.keys() if key.startswith('.')]
-
-        style = {}
-        style.update(self.style)
-        if len(keys) > 0:
-            style.update(parse_style_str(keys[0][1:], self.p.fonts))
-        element_style = process_style(element.get('style'), self.p.pdf)
-        style.update(element_style)
-
-        self.update_dimensions(style)
-
         if len(keys) > 0 or 'paragraph' in element:
             return self.process_text(element, style, element_style)
         elif 'image' in element:
@@ -733,9 +821,12 @@ class PDFContentPart:
             return self.process_table(element, style, element_style)
         elif 'content' in element:
             return self.process_child(element, style, last)
+        elif 'group' in element:
+            return self.process_group(element, style)
 
     def process_text(
-        self, element: dict, style: dict, element_style: dict
+        self, element: dict, style: dict, element_style: dict,
+        add_parts: bool=True, add_top_margin: bool=True
     ) -> dict:
         """Function that tries to add a paragraph to the current column
         rectangle, and add the remainder to the delayed list
@@ -749,14 +840,18 @@ class PDFContentPart:
         Returns:
             dict: Containing instructions to the caller.
         """
+        initial_y = self.y
+        if not self.starting and add_top_margin:
+            self.y -=  style.get('margin_top', 0)
+
+        self.update_dimensions(style)
+
         if 'paragraph' in element:
             pdf_text = element['paragraph']
-            pdf_text.setup(
-                self.x, self.y, self.width, self.max_height,
-                element['last_part'], element['last_word']
-            )
-            pdf_text.pdf = self.p.pdf
-            remaining = copy.deepcopy(element)
+            pdf_text.setup(self.x, self.y, self.width, self.max_height)
+            pdf_text.set_state(**element['state'])
+            pdf_text.finished = False
+            remaining = copy(element)
         else:
             par_style = {
                 v: style.get(v) for v in PARAGRAPH_PROPERTIES if v in style
@@ -772,18 +867,27 @@ class PDFContentPart:
         remaining['last_part'] = pdf_text.last_part
         remaining['last_word'] = pdf_text.last_word
         result['type'] = 'paragraph'
-        self.p.parts.append(result)
-        self.y -= pdf_text.current_height
+
+        if add_parts:
+            self.p.parts.append(result)
 
         if pdf_text.current_height > 0:
-            self.add_top_margin(style)
+            self.y -= pdf_text.current_height
+            self.starting = False
+            self.last_bottom = style.get('margin_bottom', 0)
+        else:
+            self.y = initial_y
 
         if pdf_text.finished:
             return {'delayed': None, 'next': False}
         else:
+            remaining['state'] = pdf_text.get_state()
             return {'delayed': remaining, 'next': True}
 
-    def process_image(self, element: dict, style: dict) -> dict:
+    def process_image(
+        self, element: dict, style: dict, add_parts: bool=True,
+        add_top_margin: bool=True
+    ) -> dict:
         """Function that tries to add an image to the current column rectangle,
         and add it to the delayed list if it can't add it.
 
@@ -794,31 +898,63 @@ class PDFContentPart:
         Returns:
             dict: Containing instructions to the caller.
         """
+        initial_y = self.y
+        if not self.starting and add_top_margin:
+            self.y -=  style.get('margin_top', 0)
+
+        self.update_dimensions(style)
+
         ret = {'delayed': None, 'next': False}
         pdf_image = PDFImage(
             element['image'], element.get('extension'),
             element.get('image_name')
         )
-        height = self.width * pdf_image.height / pdf_image.width
+        width = self.width
+        height = width * pdf_image.height / pdf_image.width
+        x = self.x
+        min_height = style.get('min_height', float('inf'))
 
+        can_add = False
         if height < self.max_height:
-            self.p.parts.append({
-                'pdf_image': pdf_image, 'type': 'image', 'x': self.x,
-                'y': self.y - height, 'width': self.width, 'height': height
-            })
+            can_add = True
+        elif min_height <= self.max_height:
+            can_add = True
+            height = min_height if style.get('shrink', 0) else self.max_height
+            new_width = height * pdf_image.width / pdf_image.height
+            x += (width - new_width) / 2
+            width = new_width
+
+        if can_add:
             self.y -= height
-            self.add_top_margin(style)
+            self.starting = False
+            self.last_bottom = style.get('margin_bottom', 0)
+            if add_parts:
+                self.p.parts.append({
+                    'pdf_image': pdf_image, 'type': 'image', 'x': x,
+                    'y': self.y, 'width': width, 'height': height
+                })
         else:
+            self.y = initial_y
+            if element.setdefault('tries', 0) >= 50:
+                raise Exception(
+                    'Image element could not be fitted in the document (try '
+                    'adding "min_height" style property to this image for us'
+                    ' to know how much we can downsize the image for it to fit)'
+                    ': {}'
+                    .format(element)
+                )
+            element['tries'] += 1
             image_place = style.get('image_place', 'flow')
             ret['delayed'] = element
             if image_place == 'normal':
                 ret['next'] = True
             elif image_place == 'flow':
-                ret['image_flow'] = True
+                ret['flow'] = True
         return ret
 
     def process_table(
-        self, element: dict, style: dict, element_style: dict
+        self, element: dict, style: dict, element_style: dict,
+        add_parts: bool=True, add_top_margin: bool=True
     ) -> dict:
         """Function that tries to add a table to the current column rectangle,
         and add the remainder to the delayed list.
@@ -832,11 +968,18 @@ class PDFContentPart:
         Returns:
             dict: Containing instructions to the caller.
         """
+        initial_y = self.y
+        if not self.starting and add_top_margin:
+            self.y -=  style.get('margin_top', 0)
+
+        self.update_dimensions(style)
+
         if 'table_delayed' in element:
             pdf_table = element['table_delayed']
             pdf_table.setup(self.x, self.y, self.width, self.max_height)
-            pdf_table.pdf = self.p.pdf
-            remaining = element
+            pdf_table.set_state(**element['state'])
+            pdf_table.finished = False
+            remaining = copy(element)
         else:
             table_props = {
                 v: element.get(v) for v in TABLE_PROPERTIES
@@ -850,20 +993,28 @@ class PDFContentPart:
             remaining = {'table_delayed': pdf_table, 'style': element_style}
 
         pdf_table.run()
-        self.p.parts.extend(pdf_table.parts)
-        self.p.lines.extend(pdf_table.lines)
-        self.p.fills.extend(pdf_table.fills)
 
-        self.y -= pdf_table.current_height
+        if add_parts:
+            self.p.parts.extend(pdf_table.parts)
+            self.p.lines.extend(pdf_table.lines)
+            self.p.fills.extend(pdf_table.fills)
+
         if pdf_table.current_height > 0:
-            self.add_top_margin(style)
+            self.y -= pdf_table.current_height
+            self.starting = False
+            self.last_bottom = style.get('margin_bottom', 0)
+        else:
+            self.y = initial_y
 
         if not pdf_table.finished:
+            remaining['state'] = pdf_table.get_state()
             return {'delayed': remaining, 'next': True}
         else:
             return {'delayed': None, 'next': False}
 
-    def process_child(self, element: dict, style: dict, last: bool) -> StrOrDict:
+    def process_child(
+        self, element: dict, style: dict, last: bool, add_top_margin: bool=True
+    ) -> StrOrDict:
         """Function that tries to add a child content to the current column
         rectangle.
 
@@ -876,9 +1027,15 @@ class PDFContentPart:
         Returns:
             str, dict: Containing instructions to the caller.
         """
+        initial_y = self.y
+        if not self.starting and add_top_margin:
+            self.y -=  style.get('margin_top', 0)
+
+        self.update_dimensions(style)
+
         pdf_content = PDFContentPart(
             element, self.p, self.get_min_x(), self.col_width, self.y,
-            self.max_y, self, last, copy.deepcopy(style)
+            self.max_y, self, last, style.copy()
         )
         down_condition1 = len(self.children_memory) > 0 and \
             self.element_index == self.section_element_index
@@ -888,9 +1045,9 @@ class PDFContentPart:
             child = self.children_memory[-1] if down_condition1 else \
                 self.partial_children_memory[-1]
             pdf_content.section_element_index = child['index']
-            pdf_content.section_delayed = copy.deepcopy(child['delayed'])
+            pdf_content.section_delayed = copy(child['delayed'])
             pdf_content.element_index = child['index']
-            pdf_content.delayed = copy.deepcopy(child['delayed'])
+            pdf_content.delayed = copy(child['delayed'])
             pdf_content.children_memory = self.children_memory[:-1] \
                 if down_condition1 else self.partial_children_memory[:-1]
 
@@ -898,13 +1055,16 @@ class PDFContentPart:
 
         action = pdf_content.run()
 
-        current_height = pdf_content.min_y - (
-            pdf_content.max_y if pdf_content.cols_n > 1 else pdf_content.y
-        )
-        self.y -= current_height
-
-        if current_height > 0:
-            self.add_top_margin(style)
+        if action != 'partial_next':
+            current_height = pdf_content.min_y - (
+                pdf_content.max_y if pdf_content.cols_n > 1 else pdf_content.y
+            )
+            if current_height > 0:
+                self.y -= current_height
+                self.starting = False
+                self.last_bottom = style.get('margin_bottom', 0)
+            else:
+                self.y = initial_y
 
         if action in ['interrupt', 'break', 'partial_next']:
             return action
@@ -912,9 +1072,107 @@ class PDFContentPart:
             self.starting = False
             return {'delayed': None, 'next': False}
 
+    def get_element_styles(self, element: dict, inherited_style: dict):
+        keys = [key for key in element.keys() if key.startswith('.')]
+
+        style = {}
+        style.update(inherited_style)
+        if len(keys) > 0:
+            style.update(parse_style_str(keys[0][1:], self.p.fonts))
+        element_style = process_style(element.get('style'), self.p.pdf)
+        style.update(element_style)
+        return style, element_style
+
+    def process_group_element(
+        self, element: dict, inherited_style: dict, add_element: bool=False,
+        add_top_margin: bool=True, min_height: Optional[Number] = None
+    ):
+        element = self.parse_element(element)
+        style, element_style = self.get_element_styles(element, inherited_style)
+
+        if min_height is not None:
+            style['min_height'] = min_height
+
+        keys = [key for key in element.keys() if key.startswith('.')]
+        if len(keys) > 0 or 'paragraph' in element:
+            return self.process_text(
+                element, style, element_style, add_element, add_top_margin
+            )
+        elif 'image' in element:
+            style['shrink'] = True
+            return self.process_image(
+                element, style, add_element, add_top_margin
+            )
+        elif 'table' in element or 'table_delayed' in element:
+            return self.process_table(
+                element, style, element_style, add_element, add_top_margin
+            )
+        else:
+            raise Exception(
+                'Element not allowed in a group element: {}'.format(element)
+            )
+
+    def process_group(self, group_element: dict, style: dict):
+        """Function that tries to add a group element to the current column
+        rectangle, and add it to the delayed list if it can't add it. If after
+        50 tries it can not add the group element, it will throw an exception.
+
+        Args:
+            group_element (dict): The group element to be added
+            style (dict): The style of the group element, combined with the
+                style of this content element.
+
+        Returns:
+            dict: Containing instructions to the caller.
+        """
+        initial_y = self.y
+
+        if not self.starting:
+            self.y -=  style.get('margin_top', 0)
+
+        images = {}
+        images_size = 0
+
+        for i, element in enumerate(group_element['group']):
+            if isinstance(element, dict) and 'image' in element:
+                image_style, _ = self.get_element_styles(element, style)
+                if 'min_height' in image_style:
+                    min_height = image_style['min_height']
+                    images[i] = min_height
+                    images_size += min_height
+
+            ans = self.process_group_element(
+                element, style, add_element=False, add_top_margin=i != 0
+            )
+            if not (isinstance(ans, dict) and ans.get('delayed') is None):
+                if group_element.setdefault('tries', 0) >= 50:
+                    raise Exception(
+                        'Group element could not be fitted in the document: {}'
+                        .format(group_element)
+                    )
+                group_element['tries'] += 1
+                self.y = initial_y
+                return {'delayed': group_element, 'next': False, 'flow': True}
+
+        new_images_size = images_size + self.y - self.max_y
+        image_ratio = new_images_size / images_size if len(images) else 1
+        self.y = initial_y
+
+        for i, element in enumerate(group_element['group']):
+            min_height = None
+            if i in images:
+                min_height = images[i]
+                min_height *= 1 if style.get('shrink', 0) else image_ratio
+            self.process_group_element(
+                element, style, add_element=True, add_top_margin=i != 0,
+                min_height=min_height
+            )
+
+        return {'delayed': None, 'next': False}
+
 from .fonts import PDFFonts
 from .image import PDFImage
 from .pdf import PDF
 from .text import PDFText
 from .table import PDFTable
-from .utils import parse_style_str, process_style
\ No newline at end of file
+from .utils import parse_style_str, process_style, copy
\ No newline at end of file
