@@ -1,13 +1,13 @@
-from copy import deepcopy
+from collections import defaultdict
 from typing import Any, Iterable, Union
 
+from pdfme.utils import parse_range_string
+
 STYLE_PROPS = dict(
     f='font_family', s='font_size', c='font_color', text_align='text_align',
     line_height='line_height', indent='indent'
 )
 
-MARGINS = ('margin_top', 'margin_right', 'margin_bottom', 'margin_left')
-
 PAGE_PROPS = ('page_size', 'rotate_page', 'margin')
 PAGE_NUMBERING = ('page_numbering_offset', 'page_numbering_style')
 
@@ -29,10 +29,8 @@ class PDFDocument:
       with all of the keys that a content box can have (see
       :class:`pdfme.content.PDFContent` for more information about content
       box, and for the default values of the attributes of this dict see
-      :class:`pdfme.pdf.PDF`).
-
-    * ``page_style``: a dict with the default attributes for each page in this
-      document. You can include any of the following keys: ``page_size``,
+      :class:`pdfme.pdf.PDF`). Additional to the keys of content box style, you
+      can add the following keys: ``outlines_level``, ``page_size``,
       ``rotate_page``, ``margin``, ``page_numbering_offset`` and
       ``page_numbering_style``. For more information about this page attributes
       and their default values see :class:`pdfme.pdf.PDF` definition.
@@ -64,6 +62,17 @@ class PDFDocument:
       height minus the bottom margin, and if it is not defined or is None i
       will be 0.
 
+    * ``per_page``: a list of dicts, each with a mandatory key ``pages``, a
+      comma separated string of indexes or ranges (python style), and any of the
+      following optional keys:
+
+      * ``style``: a style dict with page related style properties (page_size,
+        rotate_page, margin) that will be applied to every page in the ``pages``
+        ranges.
+      * ``running_sections``: a dict with optional ``exclude`` and ``include``
+        lists of running sections names to be included and excluded in every
+        page in the ``pages`` ranges.
+
     * ``sections``: an iterable with the sections of the document.
 
     Each section in ``sections`` iterable is a dict like the one that can be
@@ -72,11 +81,10 @@ class PDFDocument:
     all the contents of every section (content box) to the PDF document.
     
     Additional to the keys from a content box dict, you can
-    include also a ``page_style`` dict to overwrite the default ``page_style``
-    of the document, and a ``running_sections`` list with the name of the
+    include a ``running_sections`` list with the name of the
     running sections that you want to be included in all of the pages of the
     section. There is a special key that you can include in a section's
-    ``page_style`` dict called ``page_numbering_reset``, that if True, resets
+    ``style`` dict called ``page_numbering_reset``, that if True, resets
     the numbering of the pages.
 
     You can also include footnotes in any paragraph, by adding a dict with the
@@ -92,8 +100,8 @@ class PDFDocument:
         from pdfme import build_pdf
 
         document = {
-            "page_style": {"page_size": "letter", "margin": [70, 60]},
             "style": {
+                "page_size": "letter", "margin": [70, 60],
                 "s": 10, "c": 0.3, "f": "Times", "text_align": "j",
                 "margin_bottom": 10
             },
@@ -115,7 +123,7 @@ class PDFDocument:
             "sections": [
                 {
                     "running_sections": ["header", "footer"],
-                    "page_style": {"margin": 60},
+                    "style": {"margin": 60},
                     "content": [
                         {".": "This is a title", "style": "title"},
                         {".": [
@@ -131,7 +139,7 @@ class PDFDocument:
                 },
                 {
                     "running_sections": ["footer"],
-                    "page_style": {"rotate_page": True},
+                    "style": {"rotate_page": True},
                     "content": [
                         "This is a rotated page"
                     ]
@@ -149,32 +157,41 @@ class PDFDocument:
     """
     def __init__(self, document: dict, context: dict=None) -> None:
         context = {} if context is None else context
-        style = deepcopy(document.get('style', {}))
+        style = copy(document.get('style', {}))
         style_args = {
-            v: style.pop(k) for k, v in STYLE_PROPS.items() if k in style
+            v: style[k] for k, v in STYLE_PROPS.items() if k in style
         }
 
-        self.style_args = deepcopy(style_args)
-        self.style_args.update({k: style[k] for k in MARGINS if k in style})
-
-        page_style = document.get('page_style', {})
-        self.page_args = {
-            k: page_style[k] for k in PAGE_PROPS + PAGE_NUMBERING
-            if k in page_style
+        page_args = {
+            k: style[k] for k in PAGE_PROPS + PAGE_NUMBERING if k in style
         }
 
-        self.pdf = PDF(**self.page_args, **style_args)
+        self.pdf = PDF(
+            outlines_level=style.get('outlines_level', 1),
+            **page_args, **style_args
+        )
+
+        self.style = style
+
+        self.pdf.context.update(context)
+
         self.pdf.formats = {}
         self.pdf.formats['$footnote'] = {'r': 0.5, 's': 6}
         self.pdf.formats['$footnotes'] = {'s': 10, 'c': 0}
         self.pdf.formats.update(document.get('formats', {}))
-        self.pdf.context.update(context)
 
         self.running_sections = document.get('running_sections', {})
-        self.style = style
+
+        self.per_page = []
+        for range_dict in document.get('per_page', []):
+            new_range_dict = copy(range_dict)
+            new_range_dict['pages'] = parse_range_string(range_dict['pages'])
+            self.per_page.append(new_range_dict)
 
         self.sections = document.get('sections', [])
 
+        self.x = self.y = self.width = self.height = 0
+
         self.footnotes = []
         self._traverse_document_footnotes(self.sections)
 
@@ -223,7 +240,10 @@ class PDFDocument:
                     if isinstance(value, (list, tuple, dict)):
                         self._traverse_document_footnotes(value)
 
-    def _set_running_sections(self, running_sections: Iterable) -> None:
+    def _set_running_sections(
+        self, running_sections: Iterable, page_width: 'Number',
+        page_height: 'Number', margin: dict
+    ):
         """Method to set the running sections for every section in the document.
 
         Args:
@@ -232,12 +252,7 @@ class PDFDocument:
         """
         self.pdf.running_sections = []
         for name in running_sections:
-            section = deepcopy(self.running_sections[name])
-            margin = self.pdf.margin
-
-            page_width, page_height = self.pdf.page_width, self.pdf.page_height
-            if self.pdf.rotate_page:
-                page_width, page_height = page_height, page_width
+            section = copy(self.running_sections[name])
 
             if section.get('width') in ['left', 'right']:
                 section['width'] = margin[section.get('width')]
@@ -278,62 +293,87 @@ class PDFDocument:
         Args:
             section (dict): a dict representing the section to be processed.
         """
-        page_style = section.get('page_style', {})
-        page_args_ = {k: page_style[k] for k in PAGE_PROPS if k in page_style}
-        page_args = deepcopy(self.page_args)
-        page_args.update(page_args_)
-        self.pdf.setup_page(**page_args)
-
-        running_sections = section.get('running_sections', [])
-        self._set_running_sections(running_sections)
-
-        if 'page_numbering_offset' in page_style:
-            self.pdf.page_numbering_offset = page_style['page_numbering_offset']
-        if 'page_numbering_style' in page_style:
-            self.pdf.page_numbering_style = page_style['page_numbering_style']
-        if page_style.get('page_numbering_reset', False):
+        section_style = copy(self.style)
+        section_style.update(process_style(section.get('style', {}), self.pdf))
+        
+        if 'page_numbering_offset' in section_style:
+            self.pdf.page_numbering_offset = section_style['page_numbering_offset']
+        if 'page_numbering_style' in section_style:
+            self.pdf.page_numbering_style = section_style['page_numbering_style']
+        if section_style.get('page_numbering_reset', False):
             self.pdf.page_numbering_offset = -len(self.pdf.pages)
 
-        self.pdf.add_page()
-
-        pdf = self.pdf
-        self.width = pdf.page.width - pdf.margin['right'] - pdf.margin['left']
-        self.y = pdf.page.height - pdf.margin['top']
-        self.height = self.y - pdf.margin['bottom']
-        self.x = pdf.margin['left']
-
-        section_style = deepcopy(self.style_args)
-        section_style.update(process_style(section.get('style', {}), self.pdf))
         section['style'] = section_style
 
         self.section = self.pdf._create_content(
             section, self.width, self.height, self.x, self.y
         )
-        self._add_content()
-        while not self.section.finished:
+
+        section_page_args = {
+            k: section_style[k] for k in PAGE_PROPS if k in section_style
+        }
+
+        while True:
+            page_n = len(self.pdf.pages)
+
+            page_args = section_page_args.copy()
+
+            running_sections = set(section.get('running_sections', []))
+
+            for range_dict in self.per_page:
+                if page_n in range_dict['pages']:
+                    if 'style' in range_dict:
+                        page_style = range_dict['style']
+                        page_args.update({
+                            k: page_style[k] for k in PAGE_PROPS
+                            if k in page_style
+                        })
+                    if 'running_sections' in range_dict:
+                        per_page_rs = range_dict['running_sections']
+                        running_sections -= set(per_page_rs.get('exclude', []))
+                        running_sections.update(
+                            set(per_page_rs.get('include', []))
+                        )
+
+            self.pdf.setup_page(**page_args)
+
+            page_width, page_height = self.pdf.page_width, self.pdf.page_height
+            if self.pdf.rotate_page:
+                page_width, page_height = page_height, page_width
+
+            self.x = self.pdf.margin['left']
+            self.width = page_width - self.pdf.margin['right'] - self.x
+            self.y = page_height - self.pdf.margin['top']
+            self.height = self.y - self.pdf.margin['bottom']
+
+            self.section.setup(self.x, self.y, self.width, self.height)
+
+            self._set_running_sections(
+                running_sections, page_width, page_height, self.pdf.margin
+            )
+
             self.pdf.add_page()
+
             self._add_content()
 
+            if self.section.finished:
+                break
+
     def _add_content(self) -> None:
         """Method to add the section contents to the current page.
 
         Raises:
             Exception: if the footnotes added to the page are very large.
         """
-        content_part = self.section.pdf_content_part
-        first_page = content_part is None
-        if first_page:
-            section_element_index = 0
-            section_delayed = []
-            children_memory = []
-        else:
-            section_element_index = deepcopy(content_part.section_element_index)
-            section_delayed = deepcopy(content_part.section_delayed)
-            children_memory = deepcopy(content_part.children_memory)
+
+        section_state = self.section.get_state() \
+            if self.section.pdf_content_part is not None else {
+                'section_element_index': 0,
+                'section_delayed': [],
+                'children_memory': []
+            }
 
         self.section.run(height=self.height)
-        if first_page:
-            content_part = self.section.pdf_content_part
         footnotes_obj = self._process_footnotes()
 
         if footnotes_obj is None:
@@ -349,16 +389,16 @@ class PDFDocument:
             new_height = self.height - footnotes_obj.current_height \
                 - self.footnotes_margin
 
-            content_part.section_element_index = section_element_index
-            content_part.section_delayed = section_delayed
-            content_part.children_memory = children_memory
-
+            if section_state is not None:
+                self.section.set_state(**section_state)
+                self.section.finished = False
             self.pdf._content(self.section, height=new_height)
 
             footnotes_obj = self._process_footnotes()
 
             if footnotes_obj is not None:
                 self.pdf.page._y = self.pdf.margin['bottom'] + footnotes_height
+                self.pdf.page.x = self.x
 
                 x_line = round(self.pdf.page.x, 3)
                 y_line = round(self.pdf.page._y + self.footnotes_margin/2, 3)
@@ -408,7 +448,7 @@ class PDFDocument:
         """
         content = {'style': '$footnotes', 'content': []}
         for index, footnote in enumerate(page_footnotes):
-            footnote = deepcopy(footnote)
+            footnote = copy(footnote)
             style = footnote.setdefault('style', {})
             style.update(dict(
                 list_text=str(index + 1) + ' ', list_style='$footnote'
@@ -456,6 +496,6 @@ def build_pdf(document: dict, buffer: Any, context: dict=None) -> None:
     doc.run()
     doc.output(buffer)
 
-from .content import PDFContent
+from .content import Number, PDFContent
 from .pdf import PDF
-from .utils import process_style
\ No newline at end of file
+from .utils import process_style, copy
\ No newline at end of file
