@@ -1,4 +1,3 @@
-from copy import deepcopy
 from typing import Iterable, Optional, Union
 
 PARAGRAPH_PROPS = (
@@ -58,14 +57,15 @@ class PDFTable:
     * ``border_style``: the style of all the borders in the table. It can be
       ``solid``, ``dotted`` or ``solid``. Default value is ``solid``.
 
-    You can override the default values for the cell fills and the borders with
+    You can overwrite the default values for the cell fills and the borders with
     ``fills`` and ``borders`` arguments.
     These arguments, if passed, should be iterables of dicts. Each dict should
-    have a ``pos`` key that contains a string with information of the horizontal
-    and vertical position of the fills or borders you want to change, and so
-    this string has 2 parts separated by a semi colon, the first one for the
-    horizontal position and the second one for the vertical position. The
-    position can be a single int, a comma-separated list of ints, or a slice
+    have a ``pos`` key that contains a string with information of the vertical
+    (rows) and horizontal (columns) position of the fills or borders you want
+    to change, and for this, such a string should has 2 parts separated by a
+    semi colon, the first one for the vertical position and the second one for
+    the horizontal position.
+    The position can be a single int, a comma-separated list of ints, or a slice
     (range), like the one you pass to get a slice of a python list. For borders
     you have to include a ``h`` or a ``v`` before the positions, to tell if you
     want to change vertical or horizontal borders. The indexes in this string
@@ -73,11 +73,11 @@ class PDFTable:
 
     The following are examples of valid ``pos`` strings:
 
-    * ``'h:;0,1,-1'`` to modify the first, second and last horizontal lines in
+    * ``'h0,1,-1;:'`` to modify the first, second and last horizontal lines in
       the table. The horizontal position is a single colon, and thus the whole
       horizontal lines are affected.
 
-    * ``':;::2'`` to modify all of the fills horizontally, every two rows. This
+    * ``'::2;:'`` to modify all of the fills horizontally, every two rows. This
       would set the current fill to all the cells in the first row, the third
       row, the fifth row and so on.
 
@@ -91,7 +91,7 @@ class PDFTable:
 
     If a cell element is a dict it's ``style`` dict can have any of the
     following keys: ``cell_margin``, ``cell_margin_left``, ``cell_margin_top``,
-    ``cell_margin_right``, ``cell_margin_bottom`` and ``cell_fill``, to override
+    ``cell_margin_right``, ``cell_margin_bottom`` and ``cell_fill``, to overwrite
     the default value of any of these parameters on its cell.
     In a cell dict, you can also include ``colspan`` and ``rowspan`` keys, to
     span it horizontally and vertically respectively. The cells being merged to
@@ -218,6 +218,36 @@ class PDFTable:
         if height is not None:
             self.height = height
 
+    def get_state(self) -> dict:
+        """Method to get the current state of this table. This can be used
+        later in method :meth:`pdfme.table.PDFTable.set_state` to
+        restore this state in this table (like a checkpoint in a
+        videogame).
+
+        Returns:
+            dict: a dict with the state of this table.
+        """
+        return {
+            'current_index': self.current_index, 'delayed': copy(self.delayed)
+        }
+
+    def set_state(self, current_index: int=None, delayed: dict=None) -> None:
+        """Method to set the state of this table.
+        
+        The arguments of this method define the current state of this table,
+        and with this method you can change that state.
+
+        Args:
+            current_index (int, optional): the index of the current row being
+                added.
+            delayed (dict, optional): a dict with delayed cells that should be
+                added before the next row.
+        """
+        if current_index is not None:
+            self.current_index = current_index
+        elif delayed is not None:
+            self.delayed = delayed
+
     def set_default_border(self) -> None:
         """Method to create attribute ``default_border`` containing the default
         border values.
@@ -297,7 +327,7 @@ class PDFTable:
         self.borders_h = {}
         self.borders_v = {}
         for b in borders:
-            border = deepcopy(deepcopy(self.default_border))
+            border = copy(self.default_border)
             border.update(b)
             border['color'] = PDFColor(border['color'], True)
             pos = border.pop('pos', None)
@@ -324,9 +354,9 @@ class PDFTable:
         """
         border_l = self.borders_v if is_vert else self.borders_h
         if i in border_l and j in border_l[i]:
-            return deepcopy(border_l[i][j])
+            return copy(border_l[i][j])
         else:
-            return deepcopy(self.default_border)
+            return copy(self.default_border)
 
     def setup_fills(self, fills: Iterable) -> None:
         """Method to process the ``fills`` argument passed to this class, and
@@ -449,10 +479,12 @@ class PDFTable:
 
         can_continue = True
         if len(self.delayed) > 0:
+            can_continue = False
             row = [self.delayed.get(i) for i in range(self.cols_count)]
             action = self.add_row(row, True)
             if action == 'continue':
                 self.current_index += 1
+                can_continue = True
 
         cancel = False
         if can_continue:
@@ -634,6 +666,8 @@ class PDFTable:
                     self.fills_mem[col]['add_later'] = False
                 if self.heights_mem.get(col, 0) > self.max_height:
                     self.max_height = self.heights_mem[col]
+                if self.colspan == 0:
+                    self.is_rowspan = False
             can_continue = True
         return can_continue
 
@@ -652,7 +686,7 @@ class PDFTable:
             cell_style = element['cell_style']
             style = {}
         else:
-            style = deepcopy(self.style)
+            style = copy(self.style)
 
             if element is None:
                 element = ''
@@ -679,7 +713,7 @@ class PDFTable:
                 )
             style.pop(attr, None)
             cell_style['cell_fill'] = style.pop('cell_fill', None)
-        element = deepcopy(element)
+        element = copy(element)
         return element, style, cell_style
 
     def _setup_cell_fill(
@@ -694,11 +728,12 @@ class PDFTable:
             width (Number): the width of the cell.
             rowspan (int): the rowspan of the cell.
         """
-        fill_color = cell_style.get('cell_fill',
-            self.fills_defs[(self.current_index, col)]
-            if (self.current_index, col) in self.fills_defs
-            else None
-        )
+        fill_color = cell_style.get('cell_fill', None)
+
+        if fill_color is None:
+            fill_color = (
+                self.fills_defs.get(self.current_index, {}).get(col, None)
+            )
 
         if fill_color is not None:
             self.fills_mem[col] = { 'type': 'fill', 'x': self.x_abs,
@@ -759,24 +794,30 @@ class PDFTable:
         keys = [key for key in element.keys() if key.startswith('.')]
 
         real_height = 0
+        did_finished = False
         if len(keys) > 0 or self._is_delayed_type(element, 'text'):
-            real_height = self.process_text(
-                col, element, x, y, width, height, style, delayed
+            real_height, did_finished = self.process_text(
+                element, x, y, width, height, style, delayed
             )
         elif self.is_type(element, 'image'):
-            real_height = self.process_image(
-                col, element, x, y, width, height, delayed
+            real_height, did_finished = self.process_image(
+                element, x, y, width, height, delayed
             )
         elif self.is_type(element, 'content'):
-            real_height = self.process_content(
-                col, element, x, y, width, height, style, delayed
+            real_height, did_finished = self.process_content(
+                element, x, y, width, height, style, delayed
             )
 
         elif self.is_type(element, 'table'):
-            real_height = self.process_table(
-                col, element, x, y, width, height, style, delayed
+            real_height, did_finished = self.process_table(
+                element, x, y, width, height, style, delayed
             )
 
+        if did_finished:
+            self.delayed.pop(col, None)
+        else:
+            self.delayed[col] = delayed
+
         real_height += padd_y if real_height>0 else (0 if self.first_row else 4)
         if rowspan > 0:
             self.heights_mem[col] = real_height
@@ -787,7 +828,7 @@ class PDFTable:
         return False
 
     def process_text(
-        self, col: int, element: dict, x: Number, y: Number, width: Number,
+        self, element: dict, x: Number, y: Number, width: Number,
         height: Number, style: dict, delayed: dict
     ) -> float:
         """Method to add a paragraph to a cell.
@@ -809,7 +850,8 @@ class PDFTable:
         if 'delayed' in element:
             pdf_text = element['delayed']
             pdf_text.setup(x, y, width, height)
-            pdf_text.pdf = self.pdf
+            pdf_text.set_state(**element['state'])
+            pdf_text.finished = False
         else:
             par_style = {
                 v: style.get(v) for v in PARAGRAPH_PROPS if v in style
@@ -826,13 +868,11 @@ class PDFTable:
 
         if not pdf_text.finished:
             delayed.update({'delayed': pdf_text, 'type': 'text'})
-            self.delayed[col] = delayed
-        else:
-            self.delayed.pop(col, None)
-        return pdf_text.current_height
+            delayed['state'] = pdf_text.get_state()
+        return pdf_text.current_height, pdf_text.finished
 
     def process_image(
-        self, col: int, element: dict, x: Number, y: Number,
+        self, element: dict, x: Number, y: Number, 
         width: Number, height: Number, delayed: dict
     ) -> float:
         """Method to add an image to a cell.
@@ -858,6 +898,7 @@ class PDFTable:
         img_height = img_width * pdf_image.height / pdf_image.width
 
         real_height = 0
+        finished = False
         if img_height < height:
             real_height = img_height
             self.parts.append({
@@ -865,14 +906,13 @@ class PDFTable:
                 'x': x, 'y': y - img_height,
                 'width': img_width, 'height': img_height
             })
-            self.delayed.pop(col, None)
+            finished = True
         else:
             delayed.update({'delayed': pdf_image, 'type': 'image'})
-            self.delayed[col] = delayed
-        return real_height
+        return real_height, finished
 
     def process_content(
-        self, col: int, element: dict, x: Number, y: Number,
+        self, element: dict, x: Number, y: Number,
         width: Number, height: Number, style: dict, delayed: dict
     ) -> float:
         """Method to add a content box to a cell.
@@ -894,7 +934,8 @@ class PDFTable:
         if 'delayed' in element and element['type'] == 'content':
             pdf_content = element['delayed']
             pdf_content.setup(x, y, width, height)
-            pdf_content.pdf = self.pdf
+            pdf_content.set_state(**element['state'])
+            pdf_content.finished = False
         else:
             element['style'] = style
             pdf_content = PDFContent(
@@ -909,14 +950,11 @@ class PDFTable:
 
         if not pdf_content.finished:
             delayed.update({'delayed': pdf_content, 'type': 'content'})
-            self.delayed[col] = delayed
-        else:
-            self.delayed.pop(col, None)
-
-        return pdf_content.current_height
+            delayed['state'] = pdf_content.get_state()
+        return pdf_content.current_height, pdf_content.finished
 
     def process_table(
-        self, col: int, element: dict, x: Number, y: Number,
+        self, element: dict, x: Number, y: Number,
         width: Number, height: Number, style: dict, delayed: dict
     ) -> float:
         """Method to add a table to a cell.
@@ -938,7 +976,8 @@ class PDFTable:
         if 'delayed' in element and element['type'] == 'table':
             pdf_table = element['delayed']
             pdf_table.setup(x, y, width, height)
-            pdf_table.pdf = self.pdf
+            pdf_table.set_state(**element['state'])
+            pdf_table.finished = False
         else:
             table_props = {
                 v: element.get(v) for v in TABLE_PROPS if v in element
@@ -956,11 +995,8 @@ class PDFTable:
 
         if not pdf_table.finished:
             delayed.update({'delayed': pdf_table, 'type': 'table'})
-            self.delayed[col] = delayed
-        else:
-            self.delayed.pop(col, None)
-
-        return pdf_table.current_height
+            delayed['state'] = pdf_table.get_state()
+        return pdf_table.current_height, pdf_table.finished
 
 from .color import PDFColor
 from .content import PDFContent
@@ -968,4 +1004,4 @@ from .fonts import PDFFonts
 from .image import PDFImage
 from .pdf import PDF
 from .text import PDFText
-from .utils import parse_style_str, process_style
\ No newline at end of file
+from .utils import parse_style_str, process_style, copy
\ No newline at end of file
