@@ -1,6 +1,7 @@
+import json
 import re
-from copy import deepcopy
 from typing import Optional, Union
+from uuid import uuid4
 
 PARAGRAPH_DEFAULTS = {'text_align': 'l', 'line_height': 1.1, 'indent': 0}
 TEXT_DEFAULTS = {'f': 'Helvetica', 'c': 0.1, 's': 11, 'r': 0, 'bg': None}
@@ -518,9 +519,39 @@ class PDFTextBase:
             used_fonts=self.used_fonts, ids=self.ids,
         )
 
+    def get_state(self) -> dict:
+        """Method to get the current state of this paragraph. This can be used
+        later in method :meth:`pdfme.text.PDFText.set_state` to
+        restore this state in this paragraph (like a checkpoint in a
+        videogame).
+
+        Returns:
+            dict: a dict with the state of this paragraph.
+        """
+        return {
+            'last_part': self.last_part,
+            'last_word': self.last_word
+        }
+
+    def set_state(self, last_part: int=None, last_word: int=None) -> None:
+        """Function to update the state of the paragraph
+
+        The arguments of this method define the current state of this paragraph,
+        and with this method you can change that state.
+
+        Args:
+            last_part (int): this is the index of the part that was being
+                processed the last time method ``run`` was called.
+            last_word (int): this is the index of the
+                word of the last part that was added the last time method
+                ``run`` was called.
+        """
+        self.last_part = last_part
+        self.last_word = last_word
+
     def setup(
         self, x: Number=None, y:Number=None, width: Number=None,
-        height: Number=None, last_part: int=None, last_word: int=None
+        height: Number=None
     ):
         """Function to change any or all of the parameters of the rectangle of
         the content.
@@ -549,10 +580,6 @@ class PDFTextBase:
             self.width = width
         if height is not None:
             self.height = height
-        if last_part is not None:
-            self.last_part = last_part
-        if last_word is not None:
-            self.last_word = last_word
 
     def init(self) -> None:
         """Function to reset all of the instance properties that have to be
@@ -585,7 +612,7 @@ class PDFTextBase:
 
     def run(
         self, x: Number=None, y: Number=None, width: Number=None,
-        height: Number=None, last_part: Number=None, last_word: Number=None
+        height: Number=None
     ) -> dict:
         """Function to create the data needed to add this paragraph to the PDF
         document.
@@ -600,7 +627,7 @@ class PDFTextBase:
         Returns:
             dict: The dict from the property ``result``.
         """
-        self.setup(x, y, width, height, last_part, last_word)
+        self.setup(x, y, width, height)
         self.init()
         for part_index in range(self.last_part, len(self.content)):
             part = self.content[part_index]
@@ -966,8 +993,9 @@ class PDFText(PDFTextBase):
     of this key), which we are going to refer to as the "the dot key" from here
     on. The value for the dot key is a list/tuple containing strings or more
     ``content`` dicts like the one we are describing here (you can have nested
-    ``content`` dicts), but for simplicity, you can pass a string (for non-rich
-    text) or a tuple/list with strings and more ``content`` dicts:
+    ``content`` dicts or what we call a paragraph part), but for simplicity,
+    you can pass a string (for non-rich text) or a tuple/list with strings and
+    more paragraph parts:
 
     * If ``content`` argument is a string, it will become the following:
 
@@ -1006,7 +1034,7 @@ class PDFText(PDFTextBase):
     coming from it. The children will be able to pass their own
     style parameters to their children too, and so on.
 
-    Additional to the dot key, ``content`` dicts can have the following keys:
+    Additional to the dot key, paragraph parts can have the following keys:
 
     * ``'label'``: this is a string with a unique name (there should be
       only one label with this name in the whole document) representing
@@ -1019,6 +1047,16 @@ class PDFText(PDFTextBase):
     * ``'uri'``: this is a string with a reference to a web resource,
       that will turn this text part in a link to that web page.
 
+    * ``'outline'``: an outline is a label that is shown in the outlines panel
+      of the PDF reader. This outlines show the structure of the document.
+      This attribute is a dict with the following optional keys:
+      
+      * ``level``: an optional int with the level of this outline in the
+        outlines tree. The default value is 1.
+        
+      * ``text``: an optional string to be shown in the outlines panel for this
+        outline. The default value is the contents of this part.
+
     * ``'ids'``: when method ``run`` is called, dict attr ``result`` is
       available with information to add the paragraph to the PDF, and
       within that information you'll find a key ``ids``, a dict with
@@ -1183,8 +1221,8 @@ class PDFText(PDFTextBase):
             TypeError: If the content part passed doesn't have the format
                 described in the documentation of this class.
         """
-        style = deepcopy(parent_style)
-        ids = deepcopy(ids)
+        style = parent_style.copy()
+        ids = ids.copy()
 
         if isinstance(content, str):
             content = {'.': [content]}
@@ -1201,7 +1239,9 @@ class PDFText(PDFTextBase):
         for key, value in content.items():
             if key.startswith('.'):
                 style.update(parse_style_str(key[1:], self.fonts))
-                if isinstance(value, str):
+                if isinstance(value, (int, float)):
+                    value = [str(value)]
+                elif isinstance(value, str):
                     value = [value]
                 if not isinstance(value, (list, tuple)):
                     raise TypeError(
@@ -1219,19 +1259,32 @@ class PDFText(PDFTextBase):
         if part_var is not None:
             elements = ['0']
 
-        label = content.get('label', None)
+        label = content.get('label')
         if label is not None:
             text_part['ids'].append('$label:' + label)
-        ref = content.get('ref', None)
+        ref = content.get('ref')
         if ref is not None:
             text_part['ids'].append('$ref:' + ref)
-        uri = content.get('uri', None)
+        uri = content.get('uri')
         if uri is not None:
             text_part['ids'].append('$uri:' + uri)
+        outline = content.get('outline')
+        if isinstance(outline, dict):
+            text = outline.get('text', ''.join(str(e) for e in elements))
+            level = outline.get('level', 1)
+            if label is None:
+                outline_label = str(uuid4())
+                text_part['ids'].append('$label:' + outline_label)
+            else:
+                outline_label = label
+            outline_ = {'text': text, 'level': level, 'label': outline_label}
+            text_part['ids'].append('$outline:{}'.format(json.dumps(outline_)))
 
         is_last_string = False
 
         for element in elements:
+            if isinstance(element, (int, float)):
+                element = str(element)
             if isinstance(element, str):
                 if element == '':
                     continue
@@ -1262,4 +1315,4 @@ class PDFText(PDFTextBase):
 from .color import PDFColor
 from .fonts import PDFFonts
 from .pdf import PDF
-from .utils import get_paragraph_stream, parse_style_str, process_style
+from .utils import get_paragraph_stream, parse_style_str, process_style, copy
